import types

from asyncio import sleep  # exported

from monocle.stack.eventloop import queue_task
from monocle.callback import Callback


def next_tick():
    cb = Callback()
    cb(None)
    return cb


def immediate(val):
    cb = Callback()
    cb(val)
    return cb


def delayed(seconds, val):
    cb = Callback()
    queue_task(seconds, cb, val)
    return cb



# -*- coding: utf-8 -*-
#
# by Steven Hazel

import os
import logging
import urllib.parse
import collections
import re
import urllib.request, urllib.error, urllib.parse
import time
import base64
import http.cookies
from typing import Optional

import multidict
import yarl

import asyncio
import aiohttp.web
from aiohttp import ClientTimeout, ClientSession
from aiohttp.web_request import Request

from monocle import _o, Return, log_exception, VERSION
from monocle.stack.network import Service, SSLService, ConnectionLost, Client

try:
    from monocle.stack.network import SSLClient
except:
    pass

log = logging.getLogger(__name__)


class HttpException(Exception):
    pass


class HttpHeaders(multidict.CIMultiDict):
    def get_list(self, *a, **k):
        # synonym; should be deprecated
        return getall(*a, **k)


class HttpRequest(object):
    def __init__(self, proto='HTTP/1.0', host=None, method=None,
                 uri=None, args=None, remote_ip=None, headers=None,
                 body: Optional[str] = None, body_file: Optional[bytes] = None, cookies=None):
        self.aiohttp_request: Request = None
        self.proto = proto
        self.host = host
        self.method = method
        self.uri = uri
        self.remote_ip = remote_ip
        self.headers = headers
        if body:
            self.body_file = body.encode()
        else:
            self.body_file = body_file

        self.path, _, self.query = uri.partition('?')
        self.query_args = urllib.parse.parse_qs(self.query, keep_blank_values=True)
        self.args = args

        if cookies is not None:
            self.cookies = cookies
        else:
            self.cookies = {}
            for cookie in self.headers.getall("cookie", []):
                try:
                    for name, morsel in http.cookies.BaseCookie(cookie).items():
                        self.cookies[name] = morsel.value
                except http.cookies.CookieError:
                    pass

    @classmethod
    async def from_aiohttp_request(cls, request: Request):
        self = cls(
            proto=(request.scheme.upper() +
                   "/%s.%s" % (request.version.major, request.version.minor)),
            host=request.host,
            method=request.method,
            uri=request.raw_path,
            args=request.post(),
            remote_ip=request.transport.get_extra_info('peername')[0],
            headers=request.headers,
            body_file=await request.read(),
            cookies=request.cookies)
        self.aiohttp_request = request
        return self

    def __repr__(self):
        return "<%s (%s %s %s, headers=%s)>" % (
            self.__class__.__name__, self.method, self.path, self.proto, self.headers)

    def get_basic_auth(self):
        if 'authorization' not in self.headers:
            return None, None
        auth = self.headers["authorization"]
        try:
            method, b64 = auth.split(" ")
            if method.lower() != "basic":
                return None, None
            username, password = base64.decodestring(b64).split(':', 1)
        except Exception:
            # parsing error; no valid auth
            return None, None
        return username, password

    @property
    def body(self) -> Optional[str]:
        if self.body_file:
            return self.body_file.decode()

        return None


class HttpResponse(object):
    def __init__(self, code, msg=None, headers=None, body=None, proto=None):
        self.code = code
        self.msg = msg
        self.headers = headers or HttpHeaders()
        self.body = body
        self.proto = proto or 'HTTP/1.1'

    @classmethod
    async def from_aiohttp_response(cls, response):
        self = cls(
            code=str(response.status),
            msg=response.reason,
            headers=response.headers,
            body=await response.text(),
            proto='HTTP' + "/%s.%s" % (response.version.major, response.version.minor))
        self.aiohttp_response = response
        return self


class HttpClient(object):
    DEFAULT_PORTS = {'http': 80,
                     'https': 443}

    def __init__(self,
                 scheme: str = None,
                 host: str = None,
                 port: int = None,
                 is_proxy: bool = False):
        self.scheme = scheme
        self.host = host
        self.port = port
        self.is_proxy = is_proxy
        self.session: ClientSession = None

    async def connect(self, host, port,
                      scheme='http',
                      timeout=None,  # this parameter is deprecated
                      is_proxy=False):
        """*** DEPRECATED ***
            Provide the values via constructor instead.
        """
        if not self.is_closed():
            await self.session.close()

        self.scheme = scheme
        self.host = host
        self.port = port
        self.is_proxy = is_proxy
        self.session = ClientSession(connector=aiohttp.TCPConnector())

    async def request(self, url, headers=None, method='GET', body=None):
        parts = urllib.parse.urlsplit(url)

        if self.is_proxy:
            host = parts.netloc
            path = url
        else:
            if parts.scheme and parts.scheme != self.scheme:
                raise HttpException("URL is %s but connection is %s" %
                                    (parts.scheme, self.scheme))

            host = parts.netloc
            if not host:
                host = self.host
                if self.port != self.DEFAULT_PORTS[self.scheme]:
                    host += ":%s" % self.port

            path = parts.path
            if parts.query:
                path += '?' + parts.query
        async with self.session.request(
                method,
                self.scheme + '://' + host + path,
                headers=headers,
                data=body) as resp:
            return await HttpResponse.from_aiohttp_response(resp)

    def close(self):
        asyncio.ensure_future(self.session.close())

    def is_closed(self):
        return self.session is None or self.session.closed

    @classmethod
    async def query(cls, url, headers=None, method='GET', body=None):
        async with aiohttp.request(method, url, headers=headers, data=body) as resp:
            monocle_resp = await HttpResponse.from_aiohttp_response(resp)
            return monocle_resp


# Takes a response return value like:
# "this is a body"
# 404
# (200, "this is a body")
# (200, {"headers": "here"}, "this is a body")
#
# ...and converts that to a full (code, headers, body) tuple.
def extract_response(value):
    if isinstance(value, str):
        return (200, HttpHeaders(), value)
    if isinstance(value, int):
        return (value, HttpHeaders(), "")
    if len(value) == 2:
        return (value[0], HttpHeaders(), value[1])
    return value


class HttpRouter(object):
    named_param_re = re.compile(r':([^\/\?\*\:]+)')

    def __init__(self):
        self.routes = collections.defaultdict(list)

    @classmethod
    def path_matches(cls, path, pattern):
        m = pattern.match(path)
        if not m:
            return False, None
        if not m.end() == len(path):
            # must match the whole string
            return False, None
        return True, m.groupdict()

    def mk_decorator(self, method, pattern, add_head=False):
        if not hasattr(pattern, 'match'):
            pattern = re.escape(pattern)
            pattern = pattern.replace(r'\?', '?')
            pattern = pattern.replace(r'\:', ':')
            pattern = pattern.replace(r'\_', '_')
            pattern = pattern.replace(r'\/', '/')
            pattern = pattern.replace(r'\*', '.*')
            pattern = self.named_param_re.sub(r'(?P<\1>[^/]+)', pattern)
            pattern = re.compile("^" + pattern + "$")

        def decorator(f):
            if not asyncio.iscoroutinefunction(f):
                # don't require the @_o decorator *and* the HttpServer decorator
                handler = _o(f)
            self.routes[method].append((pattern, handler))
            if add_head:
                self.routes['HEAD'].append((pattern, handler))
            return handler

        return decorator

    def get(self, pattern, add_head=True):
        return self.mk_decorator('GET', pattern, add_head=add_head)

    def post(self, pattern):
        return self.mk_decorator('POST', pattern)

    def put(self, pattern):
        return self.mk_decorator('PUT', pattern)

    def delete(self, pattern):
        return self.mk_decorator('DELETE', pattern)

    def head(self, pattern):
        return self.mk_decorator('HEAD', pattern)

    def options(self, pattern):
        return self.mk_decorator('OPTIONS', pattern)

    def patch(self, pattern):
        return self.mk_decorator('PATCH', pattern)

    def route_match(self, req):
        for pattern, handler in self.routes[req.method]:
            match, kwargs = self.path_matches(urllib.parse.unquote(req.path),
                                              pattern)
            if match:
                return handler, kwargs
        return None, None

    @_o
    def request_handler_wrapper(self, req, handler, **kwargs):
        resp = yield handler(req, **kwargs)
        yield Return(resp)

    @_o
    def handle_request(self, req):
        before = time.time()
        resp = None

        handler, kwargs = self.route_match(req)
        try:
            if handler:
                resp = yield self.request_handler_wrapper(req, handler, **kwargs)
            elif self.handler:
                resp = yield self.request_handler_wrapper(req, self.handler)
            else:
                resp = (404, {}, "")
        except Exception:
            log_exception()
            resp = (500, {}, "500 Internal Server Error")
        after = time.time()

        content_length = 0
        if len(resp) > 2:
            content_length = len(resp[2])

        log.info("[%s] %s %s %s -> %s (%s bytes, %.0fms); %s",
                 req.remote_ip,
                 req.method, req.path, req.proto,
                 resp[0], content_length, (after - before) * 1000,
                 req.headers.get('user-agent'))

        yield Return(resp)


class HttpServer(Service, HttpRouter):
    def __init__(self, port, handler=None, bindaddr="", backlog=128,
                 max_body_str_len: int = 1024 * 1024):
        HttpRouter.__init__(self)
        self.port = port
        self.bindaddr = bindaddr
        self.backlog = backlog
        self.app = aiohttp.web.Application(client_max_size=max_body_str_len)
        self.handler = handler

        async def _handler(request: Request):
            monocle_request = await HttpRequest.from_aiohttp_request(request)
            resp = await self.handle_request(monocle_request)
            status, headers, body = extract_response(resp)
            if 'content-type' not in {k.lower() for k in headers.keys()}:
                headers['Content-Type'] = 'text/html'
            headers = {(k, str(v)) for k, v in headers.items()}
            return aiohttp.web.Response(status=status, headers=headers, body=body)

        self.app.router.add_route('*', r'/{path:.*}', _handler)

    async def _add(self):
        self._server = await asyncio.get_event_loop().create_server(
            self.app.make_handler(),
            host=self.bindaddr,
            port=self.port,
            backlog=self.backlog)


class HttpsServer(SSLService, HttpRouter):
    def __init__(self, port, ssl_options, handler=None, bindaddr="", backlog=128,
                 max_body_str_len=1024 * 1024):
        HttpRouter.__init__(self)
        SSLService.__init__(self, handler, port, bindaddr=bindaddr, backlog=backlog)



import sys
import traceback

from monocle import _o, launch
from monocle.stack import eventloop


def run(fn):
    ctx = {'exit_code': 0}

    @_o
    def _run_async(ctx):
        try:
            yield fn()
        except:
            traceback.print_exc(file=sys.stdout)
            ctx['exit_code'] = 1
        finally:
            eventloop.halt()

    launch(_run_async, ctx)
    eventloop.run()
    sys.exit(ctx['exit_code'])



# -*- coding: utf-8 -*-
#
# by Steven Hazel

from collections import deque
from .callback import Callback

from monocle.stack.eventloop import queue_task
from monocle import _o, Return


# Go-style channels
class Channel(object):
    def __init__(self, bufsize=0):
        self.bufsize = bufsize
        self._msgs = deque()
        self._recv_cbs = deque()
        self._send_cbs = deque()

    @_o
    def send(self, value):
        if self._recv_cbs:
            # if there are receivers waiting, send to the first one
            rcb = self._recv_cbs.popleft()
            queue_task(0, rcb, value)
        elif len(self._msgs) < self.bufsize:
            # if there's available buffer, use that
            self._msgs.append(value)
        else:
            # otherwise, wait for a receiver
            cb = Callback()
            self._send_cbs.append(cb)
            rcb = yield cb
            queue_task(0, rcb, value)

    @_o
    def recv(self):
        # if there's buffer, read it
        if self._msgs:
            value = self._msgs.popleft()
            yield Return(value)

        # otherwise we need a sender
        rcb = Callback()
        if self._send_cbs:
            # if there are senders waiting, wake up the first one
            cb = self._send_cbs.popleft()
            cb(rcb)
        else:
            # otherwise, wait for a sender
            self._recv_cbs.append(rcb)
        value = yield rcb
        yield Return(value)


# Some ideas from diesel:

# This is really not a very good idea without limiting it to a set of
# cancelable operations...
@_o
def first_of(*a):
    cb = Callback()
    cb.called = False
    for i, c in enumerate(a):
        def cb(result, i=i):
            if isinstance(result, Exception):
                raise result
            if not cb.called:
                cb.called = True
                cb((i, result))
        c.add(cb)
    x, r = yield cb
    yield Return([(True, r) if x == i else None for i in range(len(a))])


@_o
def fire(name, value):
    if name in waits:
        cb = waits[name]
        waits.pop(name)
        cb(value)


@_o
def wait(name):
    waits.setdefault(name, Callback())
    r = yield waits[name]
    yield Return(r)

waits = {}



# -*- coding: utf-8 -*-
#
# by Steven Hazel

import asyncio
from typing import Union

from monocle import _o, Return, launch
from monocle.callback import Callback
from monocle.stack import eventloop


class ConnectionException(Exception):
    pass


class ConnectionTimeout(asyncio.TimeoutError, ConnectionException):
    pass


"""  *** DEPRECATED *** For generic errors, please use ConnectionException instead."""
ConnectionLost = ConnectionException


class Connection(object):
    def __init__(self, reader, writer):
        self._reader = reader
        self._writer = writer
        self.writing = False
        self.flush_cb = Callback()
        self.timeout = None
        self._closed = False

    async def read_some(self) -> bytes:
        # there's no equivalent of this "read as much as can be read
        # quickly" in the asychio streaming API -- just set a
        # reasonable limit
        try:
            return await asyncio.wait_for(self._reader.read(65536), self.timeout)
        except asyncio.TimeoutError as e:
            raise ConnectionTimeout() from e

    async def read(self, size) -> bytes:
        try:
            return await asyncio.wait_for(self._reader.readexactly(size), self.timeout)
        except asyncio.TimeoutError as e:
            raise ConnectionTimeout() from e

    async def read_until(self, s: Union[bytes, str]) -> bytes:
        if isinstance(s, str):
            s = s.encode()
        try:
            return await asyncio.wait_for(self._reader.readuntil(s), self.timeout)
        except asyncio.TimeoutError as e:
            raise ConnectionTimeout() from e

    async def readline(self) -> bytes:
        try:
            return await asyncio.wait_for(self._reader.readline(), self.timeout)
        except asyncio.TimeoutError as e:
            raise ConnectionTimeout() from e

    async def write(self, data: Union[bytes, str]) -> None:
        if isinstance(data, str):
            data = data.encode()
        self._writer.write(data)
        await self.flush()

    async def flush(self):
        return await self._writer.drain()

    def close(self):
        self._closed = True
        if self._writer:
            self._writer.close()

    def is_closed(self):
        # note that this actually doesn't return True until all data
        # has been read
        return self._closed


class Service(object):
    def __init__(self, handler, port, bindaddr="", backlog=128):
        async def _handler(reader, writer):
            connection = Connection(reader, writer)
            try:
                await launch(handler, connection)
            finally:
                writer.close()

        self._handler = _handler
        self.port = port
        self.bindaddr = bindaddr
        self.backlog = backlog
        self._server = None

    async def _add(self):
        self._server = await asyncio.start_server(
            self._handler,
            host=self.bindaddr,
            port=self.port,
            limit=104857600,
            backlog=self.backlog)
        print(self._server)

    async def stop(self):
        self._server.close()
        await self._server.wait_closed()


class SSLService(Service):

    def __init__(self, handler, ssl_options, port, bindaddr="", backlog=128):
        Service.__init__(self, handler, port, bindaddr, backlog)
        self.ssl_options = ssl_options

    def _add(self):
        cf = ssl.DefaultOpenSSLContextFactory(self.ssl_options['keyfile'],
                                              self.ssl_options['certfile'])
        self._twisted_listening_port = reactor.listenSSL(
            self.port, self.factory, cf,
            backlog=self.backlog,
            interface=self.bindaddr)


class Client(Connection):
    def __init__(self, *args, **kwargs):
        Connection.__init__(self, None, None, *args, **kwargs)

    async def connect(self, host, port):
        try:
            reader, writer = await asyncio.wait_for(asyncio.open_connection(host=host, port=port, limit=104857600),
                                                    self.timeout)
            self._reader = reader
            self._writer = writer
        except asyncio.TimeoutError as e:
            raise ConnectionTimeout(f"failed to establish connection to {host}:{port} within {self.timeout}s") from e


class SSLClient(Client):

    def __init__(self, ssl_options=None):
        if ssl_options is None:
            ssl_options = {}
        Connection.__init__(self)
        self.ssl_options = ssl_options

    def _connect_to_reactor(self, host, port, factory, timeout):
        reactor.connectSSL(host, port, factory,
                           SSLContextFactory(self.ssl_options),
                           timeout=timeout)


def add_service(service):
    return asyncio.ensure_future(service._add())
